/////////////////////////////////////////////////////////////////////////
//
// File: warjo_tertex.cpp
//
// Description:
//		Implements wj_texture_gen and wj_resource class methods.  
//

#include "warjo_tertex.h"
#include "warjo_pscript.h"

#define PI 3.14159256

//**********************************************************************
//
// Terrain_Load_Line - global void function to read a line from the text file
//
char sz_terr_buffer[512];

void Terrain_Load_Line(void)
{
	char sztmp[128];
    wj_terrain_type terr;

	//get the texture name
	if (! PureScript::GetStrN(0,sz_terr_buffer,sztmp) ) return;

	//add the texture in the resource
	terr.texID=g_resource.AddTexture(sztmp);

	//colors
	if (! PureScript::GetStrN(1,sz_terr_buffer,sztmp) ) return;
	terr.r=atof(sztmp)/255;
	if (! PureScript::GetStrN(2,sz_terr_buffer,sztmp) ) return;
	terr.g=atof(sztmp)/255;
	if (! PureScript::GetStrN(3,sz_terr_buffer,sztmp) ) return;
	terr.b=atof(sztmp)/255;

	//the elevations
	if (! PureScript::GetStrN(4,sz_terr_buffer,sztmp) ) return;
	terr.elevMin=atof(sztmp);	
	
	if (! PureScript::GetStrN(5,sz_terr_buffer,sztmp) ) return;
	terr.elevMax=atof(sztmp);	
	
	if (! PureScript::GetStrN(6,sz_terr_buffer,sztmp) ) return;
	terr.release=atof(sztmp);	
	
	//the slope
	if (! PureScript::GetStrN(7,sz_terr_buffer,sztmp) ) return;
	terr.slopeMin=atof(sztmp);	
	if (! PureScript::GetStrN(8,sz_terr_buffer,sztmp) ) return;
	terr.slopeMax=atof(sztmp);	
	if (! PureScript::GetStrN(9,sz_terr_buffer,sztmp) ) return;
	terr.sloperelease=atof(sztmp);

	//the skew
	if (! PureScript::GetStrN(10,sz_terr_buffer,sztmp) ) return;
	terr.skew=atof(sztmp);
	if (! PureScript::GetStrN(11,sz_terr_buffer,sztmp) ) return;
	terr.skewAzimuth=atof(sztmp);

	//add the terrain type to the collection
	g_resource.terrs.push_back(terr);
}

//**********************************************************************
//
// wj_resource
//
//----------------------------------------------------------------------
//
// AddTexture - it will load the bmp file pointed by szname and return
//				an ID in the textures collection
//
BYTE wj_resource::AddTexture(const char *szname)
{
	for(int i=0;i<texs.size();i++) 
		if (strcpy(texs[i].texture,szname)==0) return i;
	
	wj_texture_type tex;
	strcpy(tex.texture,szname);

	//load the bmp;
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER bmch;
	FILE *fp=fopen(szname,"rb");
	if (fp==NULL) return 0xff;
	fread(&bmfh,sizeof(bmfh),1,fp);
	fread(&bmch,sizeof(bmch),1,fp);
	tex.height=bmch.biHeight;
	tex.width=bmch.biWidth;
	tex.pb=(BYTE *)malloc(bmch.biWidth*bmch.biHeight*3);
	fread(tex.pb,1,bmch.biWidth*bmch.biHeight*3,fp);
	fclose(fp);

	texs.push_back(tex);
	return texs.size()-1;
}

//----------------------------------------------------------------------
//
// GetTexturePixel - retrieves the color information in r,g,b from texture
//					 "id" and texture coordinates tu,tv
//
void wj_resource::GetTexturePixel(BYTE id,int tu,int tv,
				float &r,float &g,float &b)
{
	BYTE *pb=texs[id].pb;
	tu%=texs[id].width;
	tv%=texs[id].height;
	DWORD dwoff=(texs[id].width*tv+tu)*3;
	b=(float)(pb[dwoff])/255.0f;
	g=(float)(pb[dwoff+1])/255.0f;
	r=(float)(pb[dwoff+2])/255.0f;
}

//**********************************************************************
//
// wj_texture_gen
//
//----------------------------------------------------------------------
//
// static members
//
float wj_texture_gen::mapscale=64;
VECTOR3 wj_texture_gen::dlight(-1,-1,-1);
int wj_texture_gen::pside=64;

//----------------------------------------------------------------------
//
// static Init function - maps static members to text file identifiers
//
void wj_texture_gen::Init(void)
{
	g_pscript.Map((char *)sz_terr_buffer,"terrain");
	g_pscript.Mapf(Terrain_Load_Line,"terrain");
	g_pscript.Map(&(dlight.x),"TrDirectLight.x");
	g_pscript.Map(&(dlight.y),"TrDirectLight.y");
	g_pscript.Map(&(dlight.z),"TrDirectLight.z");
	g_pscript.Map(&(mapscale),"mapscale");
	g_pscript.Map(&(pside),"patchside");
}

//----------------------------------------------------------------------
//
// constructor (memory allocation failure)
//
wj_texture_gen::wj_texture_gen(int w,int h):m_width(w),m_height(h)
{
	dlight.Normalize();
	slight=0.3f;
	bitmap=(BYTE *)malloc(w*h*3);
}

//----------------------------------------------------------------------
//
// destructor
//
wj_texture_gen::~wj_texture_gen()
{
	if (bitmap) free(bitmap);
}

//----------------------------------------------------------------------
//
// TRender - renders a pixel with the given h and normal, tu,tv are to be
//			 used as texture coordinates if there is texture in the 
//		  	 terrain type
//
DWORD wj_texture_gen::TRender(float h,VECTOR3 &normal,float tu,float tv)
{
	//iterator in the terrain type collection
	iter_terrain it;
	
	//color factor collection and iterator
	std::list<color_factor> cols;
	std::list<color_factor>::iterator ci;

	//total factor used for normaization
	float totalfactor=0.0f;

	//slope of the current point (the y value of the normal)
	float slope=normal.y;

	//the skew denominator
	float skewDenom=normal.x*normal.x+normal.z*normal.z;

	//are we to do skewing... if it is a flat terrain there is no need
	bool doSkew;

	//if skew denominator is "almost" 0 then terrain is flat
	if (skewDenom >0.00001) {
		//turn the skewing on and calculate the denominator
		doSkew=true;
		skewDenom=1.0f/sqrt(skewDenom);
	}
	else doSkew=false;;

	//iterate through terrain type collection
	for(it=g_resource.terrs.begin();it!=g_resource.terrs.end();it++) {

		//the new color factor
		color_factor cf;

		//factor for this material (for clarity)
		float factor=1.0; 

		//first check - elevation
		float elv_max=it->elevMax;
		float elv_min=it->elevMin;

		//are we to do skewing ?
		if (doSkew) {
			//calculate 2D skew vector
			float skx=cos(it->skewAzimuth*PI/180);
			float sky=sin(it->skewAzimuth*PI/180);

			//skew scale value
			float scale=(normal.x*skx+normal.y*sky)*skewDenom;
			
			//adjust elevation limits
			elv_max+=it->skew*scale;
			elv_min+=it->skew*scale;
		}
		//current elevation release
		float rel=it->release;

		//if outside limit elevation AND release skip this one
		if (elv_max+rel<h) continue;
		if (elv_min-rel>h) continue;
		
		if (elv_max<h) {
				// we are in release compute the factor
				factor=1-(h-elv_max)/rel;
			}
		if (elv_min>h) {
				// we are in release compute the factor
				factor=1-(elv_min-h)/rel;
			}

		//now check the slopes...

		//slope release
		float srel=cos(PI/2-(it->sloperelease*PI/180));

		//calculate min and max slope
		float minslope=cos(it->slopeMin*PI/180);
		float maxslope=cos(it->slopeMax*PI/180);

		//reverse?
		if (minslope>maxslope) {
			float t=maxslope;maxslope=minslope;minslope=t;
		}

		//this slope is not supported for this type
		if (slope<minslope-srel) continue;
		if (slope>maxslope+srel) continue;

		//release?
		if (slope>maxslope) factor*=1-(slope-maxslope)/srel;
		if (slope<minslope) factor*=1-(minslope-slope)/srel;

		//this is our terrain type - get the color info
		if (it->texID==0xff) {
			cf.r=it->r;
			cf.g=it->g;
			cf.b=it->b;
			}
		else 
		  g_resource.GetTexturePixel(it->texID,tu,tv,cf.r,cf.g,cf.b);

		//and the factor
		cf.factor=factor;

		//add to the collection
		cols.push_front(cf);

		//add it to the total factor for normalization
		totalfactor+=factor;
	}

	//compute direct light
	float light=-normal.DotProduct(dlight);
	if (light<0) light=0;

	//add static (ambient) light 
	light+=slight;

	//rgb values
	float r,g,b;
	r=g=b=0;

	//normalizing and acumulating factors
	for(ci=cols.begin();ci!=cols.end();ci++) {
		float factor=ci->factor/totalfactor;
		r+=ci->r*factor;
		g+=ci->g*factor;
		b+=ci->b*factor;
	}

	//light applied
	r*=light;
	g*=light;
	b*=light;

	//return the pixel color
	return RGB(b*255,g*255,r*255);
}

//----------------------------------------------------------------------
//
// MakeTexture - generate texture from "map" bitmap with dimesions (w,h)
//				 and offset of the patch (xf,yf)
//
void wj_texture_gen::MakeTexture(BYTE *map,int w,int h,int xf,int yf)
{
	//map interpolation object
	MapInter mi(map,w,h,mapscale);

	//step 
	float dx=(float)(pside)/(float)m_width;
	float dy=(float)(pside)/(float)m_height;

	//normalize directional light
	dlight.Normalize();

	//iterate through every texel in the texture 
	for(int y=0;y<m_height;y++)
		for(int x=0;x<m_width;x++) {
			
			//real coordinates
			float tr_x=(float)x*dx+xf;
			float tr_y=(float)y*dy+yf;

			//compute the height
			float h=mi.GetHeight(tr_x,tr_y);

			//calculate the normal... tt
			VECTOR3 tt(0,0,0);
			{
			VECTOR3 dv1(0,mi.GetHeight(dx*(float)x+xf,dy*(float)(y-1)+yf)-mi.GetHeight(dx*(float)x+xf,dy*(float)(y+1)+yf),dy*2);
			VECTOR3 dv2(dx*2,mi.GetHeight(dx*(float)(x-1)+xf,dy*(float)(y)+yf)-mi.GetHeight(dx*(float)(x+1)+xf,dy*(float)(y)+yf),0);
			tt.CrossProduct(dv1,dv2);
			tt.Normalize();
			}

			//render the pixel
			DWORD dwCol=TRender(h,tt,tr_x,tr_y);

			//put it in the bitmap
			memmove(bitmap+(y*m_width+x)*3,&dwCol,3);
		}
}

//----------------------------------------------------------------------
//
// Mirror - the texture. mx and my are flags... if mx is true then mirror
//			in x coordinate. if my is true mirror in y coordinate			
//
void wj_texture_gen::Mirror(BYTE mx,BYTE my)
{
	//both false then no mirror
	if ((!mx)&&(!my)) return;

	//mirror - very bad method
	int maxy=(my)?m_height/2:m_height;
	int maxx=(my)?m_width:((mx)?m_width/2:m_width);

	for(int y=0;y<maxy;y++)
		for(int x=0;x<maxx;x++) {
			int nx=x;
			int ny=y;

			if (mx) 
				nx=m_width-1-nx;
			if (my) 
				ny=m_height-1-ny;

			int of=(x+y*m_width)*3;
			int nof=(nx+ny*m_width)*3;

			DWORD dwt;
			memmove(&dwt,bitmap+of,3);
			memmove(bitmap+of,bitmap+nof,3);
			memmove(bitmap+nof,&dwt,3);
		}

}